
//
//
// Project       : Vorpal-Mail
//
// Source        : virtual.cc
//
// Creator       : Blane Bramble                    Date          : 27th October 2000
// Amended       : Blane Bramble                    Date          : 7th December 2000
//
// License       : Copyright Virtua-Web (to be released under GPL)
//
// virtual.cc provides the classes required to read in a Vorpal-Mail
// configuration file, and to expand and verify user names using
// that data.
//

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "mail.h"

#include "list.h"
#include "hash.h"
#include "array.h"
#include "virtual.h"
#include "smtp.h"

char *VIRTUAL_DOMAIN::keywords[] = {
	"domain",
	"smtprelay",
	"masquerade",
	"relay",
	"user",
	"list",
	"virtual",
	"queuein",
	"queueout",
	"unknown",
	"local",
	"scanner",
	"port",
	NULL,
	};

enum {
	VIRTDOM_KEY_DOMAIN = 0,
	VIRTDOM_KEY_SMTPRELAY,
	VIRTDOM_KEY_MASQUERADE,
	VIRTDOM_KEY_RELAY,
	VIRTDOM_KEY_USER,
	VIRTDOM_KEY_LIST,
	VIRTDOM_KEY_VIRTUAL,
	VIRTDOM_KEY_QUEUEIN,
	VIRTDOM_KEY_QUEUEOUT,
	VIRTDOM_KEY_UNKNOWN,
	VIRTDOM_KEY_LOCAL,
	VIRTDOM_KEY_SCANNER,
	VIRTDOM_KEY_PORT,
	};

//
// Private class functions
//
char *VIRTUAL_DOMAIN::getword(char *from, char *to)
{
int quoted = 0, empty = 1;
*to = 0;
while(isspace(*from))
	from++;

while(*from != 0)
	{
	if (quoted)
		{
		if (*from == '"')
			{
			quoted = 0;
			from++;
			}
		else
			*to++ = *from++;
		}
	else
		{
		if (isspace(*from))
			break;
		else if (*from == '#')
			{
			from += strlen(from);
			break;
			}
		else if (*from == '"')
			{
			quoted = 1;
			from++;
			}
		else if (*from == ',' && !empty)
			break;
		else
			*to++ = *from++;
		empty = 0;
		}
	}
*to = 0;
return from;
}

int VIRTUAL_DOMAIN::lookup(char *word)
{
int n = 0;
while(keywords[n] != NULL)
	{
	if (strcasecmp(word, keywords[n]) == 0)
		return n;
	n++;
	}
return -1;
}

VIRTUAL_DOMAIN *VIRTUAL_DOMAIN::finddomain(char *dom)
{
int n;

if (*dom == 0 || strcasecmp(this->domainname, dom) == 0)
	return this;
if (this->masquerade != NULL)
	{
	for(n=0;n<this->masquerade->getsize();n++)
		{
		if (strcasecmp(this->masquerade->get(n), dom) == 0)
			return this;
		}
	}
if (virtualdomain != NULL)
	{
	LISTITEM *item;
	item = virtualdomain->gethead();
	while(item != NULL)
		{
		VIRTUAL_DOMAIN *d;
		if ((d = ((VIRTUAL_DOMAIN *)(item->getdata()))->finddomain(dom)) != NULL)
			return d;
		item = item->getnext();
		}
	}
return NULL;
}

VIRTUAL_DOMAIN::VIRTUAL_DOMAIN(void)
{
listenport = 25;
domainname = NULL;
smtprelay = NULL;
local = NULL;
masquerade = new STRING_ARRAY();
userlist = new HASHTABLE(1013);
listlist = new HASHTABLE(1013);
relaylist = new LIST();
status = VIRTDOM_KEYWORD;
scratch = NULL;
scanner = NULL;
virtualdomain = new LIST();
queuein = strdup("/var/spool/vmail/in");
queueout = strdup("/var/spool/vmail/out");
unknown = VIRTDOM_UNKNOWN_REJECT;
unknownrelay = NULL;
}

VIRTUAL_DOMAIN::~VIRTUAL_DOMAIN()
{
if (scanner != NULL)
	free(scanner);
if (domainname != NULL)
	free(domainname);
if (smtprelay != NULL)
	free(smtprelay);
if (masquerade != NULL)
	delete masquerade;
if (relaylist != NULL)
	{
	LISTITEM *item;
	item = relaylist->gethead();
	while(item != NULL)
		{
		delete ((LIST *)(item->getdata()));
		item = item->getnext();
		}
	delete relaylist;
	}
if (userlist != NULL)
	delete userlist;
if (listlist != NULL)
	delete listlist;
if (scratch != NULL)
	delete scratch;
if (virtualdomain != NULL)
	{
	LISTITEM *item;
	item = virtualdomain->gethead();
	while(item != NULL)
		{
		delete ((VIRTUAL_DOMAIN *)(item->getdata()));
		item = item->getnext();
		}
	delete virtualdomain;
	}
if (unknownrelay != NULL)
	free(unknownrelay);
if (local != NULL)
	free(local);
free(queuein);
free(queueout);
}

void VIRTUAL_DOMAIN::process(char *w)
{
switch(this->status)
	{
	case VIRTDOM_KEYWORD:
		if (*w == ',')
			status = VIRTDOM_VALUE;			// it's another value!
		else if ((command = lookup(w)) == -1)
			status = VIRTDOM_KEYWORD;		// unknown keyword
		else
			status = VIRTDOM_ARGUMENT;
		break;
	case VIRTDOM_ARGUMENT:
		switch(command)
			{
			case VIRTDOM_KEY_DOMAIN:
				if (domainname == NULL)
					domainname = strdup(w);
				status = VIRTDOM_KEYWORD;
				break;
			case VIRTDOM_KEY_SCANNER:
				if (scanner == NULL)
					scanner = strdup(w);
				status = VIRTDOM_KEYWORD;
				break;
			case VIRTDOM_KEY_PORT:
				listenport = (int)strtol(w, NULL, 10);
				status = VIRTDOM_KEYWORD;
				break;
			case VIRTDOM_KEY_SMTPRELAY:
				if (smtprelay == NULL)
					smtprelay = strdup(w);
				status = VIRTDOM_KEYWORD;
				break;
			case VIRTDOM_KEY_MASQUERADE:
				masquerade->add(w);
				status = VIRTDOM_KEYWORD;
				break;
			case VIRTDOM_KEY_RELAY:
				{
				LIST *data;

				if ((data = new LIST()) != NULL)
					{
					data->add(w, "relay");			// add relay item to list
	
					relaylist->add(NULL, data);		// add new list to relaylist
	
					sublist = data;
					}
				status = VIRTDOM_KEYWORD;
				}
				break;
			case VIRTDOM_KEY_VIRTUAL:
				{
				VIRTUAL_DOMAIN *dom;
				dom = new VIRTUAL_DOMAIN();
				virtualdomain->add(w, (void *)dom);
				dom->load(w);
				status = VIRTDOM_KEYWORD;
				}
				break;
			case VIRTDOM_KEY_USER:
				if (userlist->find(w) == NULL)
					{
					LIST *data;
					data = new LIST();
					userlist->add(w, (void *)data);
					}
				if (scratch != NULL)
					free(scratch);
				scratch = strdup(w);
				status = VIRTDOM_VALUE;
				break;
			case VIRTDOM_KEY_LIST:
				if (listlist->find(w) == NULL)
					{
					LIST *data;
					data = new LIST();
					listlist->add(w, (void *)data);
					}
				if (scratch != NULL)
					free(scratch);
				scratch = strdup(w);
				status = VIRTDOM_VALUE;
				break;
			case VIRTDOM_KEY_QUEUEIN:
				free(queuein);
				queuein = strdup(w);
				status = VIRTDOM_KEYWORD;
				break;
			case VIRTDOM_KEY_QUEUEOUT:
				free(queueout);
				queueout = strdup(w);
				status = VIRTDOM_KEYWORD;
				break;
			case VIRTDOM_KEY_UNKNOWN:
				if (strcasecmp(w, "reject") == 0)
					unknown = VIRTDOM_UNKNOWN_REJECT;
				else if (strcasecmp(w, "accept") == 0)
					unknown = VIRTDOM_UNKNOWN_ACCEPT;
				else if (strcasecmp(w, "relay") == 0)
					unknown = VIRTDOM_UNKNOWN_RELAY;
				else
					{
					if (unknownrelay != NULL)
						free(unknownrelay);
					unknownrelay = strdup(w);
					unknown = VIRTDOM_UNKNOWN_MAILTO;
					}
				status = VIRTDOM_KEYWORD;
				break;
			case VIRTDOM_KEY_LOCAL:
				if (local != NULL)
					free(local);
				local = strdup(w);
				status = VIRTDOM_KEYWORD;
				break;
			}
		break;
	case VIRTDOM_VALUE:
		switch(command)
			{
			LISTITEM *item;
			LIST *data;
			char *ptr;

			case VIRTDOM_KEY_RELAY:
				sublist->add(w, "relay");
				status = VIRTDOM_KEYWORD;
				break;

			case VIRTDOM_KEY_USER:
#ifdef OLDMETHOD
				if ((ptr = strchr(w, ',')) != NULL)		// if ends in , (more values expected)
					*ptr = 0;				// remove ,
				else
					status = VIRTDOM_KEYWORD;		// otherwise expect a keyword next
#else
				status = VIRTDOM_KEYWORD;
#endif
				if ((item = userlist->find(scratch)) != NULL)
					{
					data = (LIST *)(item->getdata());
					data->add(w, "user");
					}
				break;
			case VIRTDOM_KEY_LIST:
#ifdef OLDMETHOD
				if ((ptr = strchr(w, ',')) != NULL)		// if ends in , (more values expected)
					*ptr = 0;				// remove ,
				else
					status = VIRTDOM_KEYWORD;		// otherwise expect a keyword next
#else
				status = VIRTDOM_KEYWORD;
#endif
				if ((item = listlist->find(scratch)) != NULL)
					{
					data = (LIST *)(item->getdata());
					data->add(w, "user");
					}
				break;
			}
		break;
	}
}

int VIRTUAL_DOMAIN::load(char *fn)
{
FILE *fp;
char linebuf[MAX_LINE];
char *ptr;
char wordbuf[MAX_LINE];

ptr = linebuf;
*ptr = 0;

if ((fp = fopen(fn, "r")) == NULL)
	return 0;

while(fgets(linebuf, MAX_LINE-1, fp) != NULL)
	{
	ptr = linebuf;
	while(*(ptr = getword(ptr, wordbuf)) != 0)
		{
		process(wordbuf);
		}
	}
fclose(fp);
}

void VIRTUAL_DOMAIN::dump(void)
{
int n;
char *str;

printf("---- DUMP ----\n");

if (smtprelay != NULL)
	printf("smtprelay = %s\n", smtprelay);
if (domainname != NULL)
	printf("domain = %s\n", domainname);
if (local != NULL)
	printf("local = %s\n", local);	

printf("unknown = %d\n", unknown);

if (masquerade != NULL)
	{
	n = 0;
	while((str = masquerade->get(n)) != NULL)
		{
		printf("masquerade[%d] = %s\n", n, str);
		n++;
		}
	}
if (relaylist != NULL)
	{
	LISTITEM *item;	
	LIST *data;

	item = relaylist->gethead();
	while(item != NULL)
		{
		LISTITEM *ritem;
		data = ((LIST *)item->getdata());
		if (data != NULL)
			{
			ritem = data->gethead();
			printf("relay ");
			while(ritem != NULL)
				{
				printf("%s, ", ritem->getkey());
				ritem = ritem->getnext();
				}
			}
		printf("\n");
		item = item->getnext();
		}
	}

if (userlist != NULL)
	{
	int n;
	LIST *list, *data;
	for(n=0;n<userlist->getsize();n++)
		{
		if ((list = userlist->get(n)) != NULL)		// get the list of lists from the hash table
			{
			LISTITEM *item;
			item = list->gethead();			// get the first list with this hash value
			while(item != NULL)			// for each hash list item (primary name)
				{
				LISTITEM *ditem;
				printf("%s = ", item->getkey());
				if ((data = (LIST *)item->getdata()) != NULL)		// get the list of secondary names
					{
					ditem = data->gethead();
					while(ditem != NULL)
						{
						printf("%s, ", ditem->getkey());
						ditem = ditem->getnext();
						}
					}
				printf("\n");
				item = item->getnext();
				}
			}
		}
	}
if (listlist != NULL)
	{
	int n;
	LIST *list, *data;
	for(n=0;n<listlist->getsize();n++)
		{
		if ((list = listlist->get(n)) != NULL)		// get the list of lists from the hash table
			{
			LISTITEM *item;
			item = list->gethead();			// get the first list with this hash value
			while(item != NULL)			// for each hash list item (primary name)
				{
				LISTITEM *ditem;
				printf("%s = ", item->getkey());
				if ((data = (LIST *)item->getdata()) != NULL)		// get the list of secondary names
					{
					ditem = data->gethead();
					while(ditem != NULL)
						{
						printf("%s, ", ditem->getkey());
						ditem = ditem->getnext();
						}
					}
				printf("\n");
				item = item->getnext();
				}
			}
		}
	}

if (virtualdomain != NULL)
	{
	LISTITEM *list;
	VIRTUAL_DOMAIN *dom;
	list = virtualdomain->gethead();
	while(list != NULL)
		{
		dom = (VIRTUAL_DOMAIN *)list->getdata();
		printf("%s:\n", list->getkey());
		dom->dump();
		list = list->getnext();
		}
	}

}

char *VIRTUAL_DOMAIN::toemail(char *from)
{
static char full[SMTP_MAX_NAME];
if (strchr(from, '@') != NULL)
	return from;
snprintf(full, SMTP_MAX_NAME - 1,"%s@%s", from, this->domainname);
return full;
}

LIST *VIRTUAL_DOMAIN::expand(char *email)
{
int hasdomain = 0;
char *domain, *name, *temp, *ptr;
VIRTUAL_DOMAIN *dom;
LIST *expanded = NULL, *list;
LISTITEM *item;

temp = strdup(email);

ptr = temp;
while(*ptr != 0)
	{
	if (*ptr == '@')
		{
		*ptr = 0;
		hasdomain = 1;
		}
	else
		ptr++;
	}
name = strdup(temp);
if (hasdomain)
	{
	*ptr = '@';
	domain = strdup(ptr+1);
	}
else
	{
	domain = strdup(this->getname());
	}
free(temp);

// At this point, name and domain are valid.

if ((dom = finddomain(domain)) == NULL)
	{
	free(name);
	free(domain);
	return NULL;
	}

listlist->untouch();
userlist->untouch();

if ((item = dom->listlist->touchfind(name)) != NULL)		// got a name in an expansion list
	{
	if ((list = (LIST *)item->getdata()) != NULL)	// get the list of expansions (shouldn't ever be null, but what the hell!
		{
		item = list->gethead();
		while(item != NULL)			// walk the valid expansions
			{
			LIST *recurse;
			if (expanded == NULL)
				expanded = new LIST;
			if ((recurse = expand(dom->toemail(item->getkey()))) != NULL)		// an item was a list!
				{
				LISTITEM *ritem;
				ritem = recurse->gethead();
				while(ritem != NULL)
					{
					expanded->add(dom->toemail(ritem->getkey()), ritem->getdata());	
					ritem = ritem->getnext();
					}
				delete recurse;
				}
			else
				expanded->add(dom->toemail(item->getkey()), item->getdata());
			item = item->getnext();
			}
		}
	}
else if ((item = dom->userlist->touchfind(name)) != NULL)		// got a name in an expansion list
	{
	if ((list = (LIST *)item->getdata()) != NULL)	// get the list of users.
		{
		item = list->gethead();
		while(item != NULL)			// walk the valid expansions
			{
			LIST *recurse;
			if (expanded == NULL)
				expanded = new LIST();
#ifdef EXPAND_USERS
			if ((recurse = expand(dom->toemail(item->getkey()))) != NULL)		// expand
				{
				LISTITEM *ritem;
				ritem = recurse->gethead();
				while(ritem != NULL)
					{
					expanded->add(dom->toemail(ritem->getkey()), ritem->getdata());
					ritem = ritem->getnext();
					}
				delete recurse;
				}
			else
#endif
				expanded->add(dom->toemail(item->getkey()), item->getdata());
			item = item->getnext();
			}
		}
	}

free(name);
free(domain);
return expanded;
}

/* User verify method:

Split into user and domain
Lookup domain in list of virtual domains, if found use that .conf, otherwise use vmail.conf (defaults)
Lookup user in list
If found, expand and lookup each expansion in list (recursively)
If not found, lookup user in user
If found, expand, attempt local delivery for any without domain. Any with domain, lookup domain in virtual domains, if found start again, else queue for remote delivery.


*/

