
//
//
// Project       : Vorpal-Mail
//
// Source        : virtual.h
//
// Creator       : Blane Bramble                    Date          : 27th October 2000
// Amended       : Blane Bramble                    Date          : 29nd November 2000
//
// License       : Copyright Virtua-Web
//
// Definitions for virtual domain configurations for Vorpal-Mail
//
//

#include "hash.h"
#include "array.h"

#ifndef __VIRTUAL_DOMAIN__
#define __VIRTUAL_DOMAIN__

#define EXPAND_USERS

enum {
	VIRTDOM_KEYWORD,
	VIRTDOM_ARGUMENT,
	VIRTDOM_VALUE
	};

enum VIRTDOM_UNKNOWN {
	VIRTDOM_UNKNOWN_REJECT,
	VIRTDOM_UNKNOWN_RELAY,
	VIRTDOM_UNKNOWN_ACCEPT,
	VIRTDOM_UNKNOWN_MAILTO,
	};

class VIRTUAL_DOMAIN {
	private:
		static char *keywords[];

		int command, status;

		char *scratch;

		char *queuein, *queueout;

		char *domainname;
		char *smtprelay;
		char *local;
		char *scanner;
		int listenport;

		STRING_ARRAY *masquerade;
		LIST *relaylist, *sublist;
		HASHTABLE *userlist;
		HASHTABLE *listlist;
		LIST *virtualdomain;

		VIRTDOM_UNKNOWN unknown;
		char *unknownrelay;

		char *getword(char *, char *);
		int lookup(char *);
		void process(char *);
	public:
		VIRTUAL_DOMAIN(void);
		~VIRTUAL_DOMAIN();
		int load(char *);
		void dump(void);
		LIST *getvirtual(void) { return virtualdomain; };
		LIST *getrelay(void) { return relaylist; };
		char *getsmtp(void) { return smtprelay; };
		LIST *expand(char *);
		char *toemail(char *);
		char *getname(void) { return domainname; };
		char *getqin(void) { return queuein; };
		char *getqout(void) { return queueout; };
		VIRTUAL_DOMAIN *finddomain(char *);
		VIRTDOM_UNKNOWN getunknown(void) { return unknown; };
		char *getunknownrelay(void) { return unknownrelay; };
		char *getlocal(void) { return local; };
		char *getscanner(void) { return scanner; };
		int getport(void) { return listenport; };
	};

#endif

//
// userlist [HASHTABLE]
//  |
// LIST
//  |
// LISTITEM (user1) -- LIST
//  |                   |
//  |                  LISTITEM (alias1)
//  |                  LISTITEM (alias2)
//  |
// LISTITEM (user2) -- LIST
//                      |
//                     LISTITEM (alias1)
//                     LISTITEM (alias2)

